#include <math.h>

#include "bitmap.h"

void line1(Bitmap<float> &image,int x1,int y1,int x2,int y2){
  for (int x=x1;x<=x2;x++){
    int y=y1+(int) floor((float) (x-x1)*(y2-y1)/(x2-x1)+0.5);
    image(x,y)=white();
  }
}

void line2(Bitmap<float> &image,int x1,int y1,int x2,int y2){
  float r=(float) (y2-y1)/(x2-x1),t=1/sqrt(1+r*r);
  for (int x=x1;x<=x2;x++)
    for (int y=y1;y<=y2;y++){
      float delta=t*fabs(y1+(float) (x-x1)*(y2-y1)/(x2-x1)-y);
      if (delta<=1)
        image(x,y)=gray(255*(1-delta));
    }
}

void copy(Bitmap<float> &source,Bitmap<float> &dest,int x,int y){
  for (int i=0;i<source.getWidth();i++)
    for (int j=0;j<source.getHeight();j++)
      dest(x+i,y+j)=source(i,j);
}

void underSample(Bitmap<float> &source,Bitmap<float> &dest){
  dest.setSize(source.getWidth()/2,source.getHeight()/2);
  for (int i=0;i<dest.getWidth();i++)
    for (int j=0;j<dest.getHeight();j++){
      dest(i,j)=source(2*i,2*j)+source(2*i+1,2*j)+source(2*i,2*j+1)+source(2*i+1,2*j+1);
      dest(i,j)*=0.25;
    }
}

void multiUnderSample(Bitmap<float> &image,int n){
  Bitmap<float> temp;
  for (int i=0;i<n;i++){
    underSample(image,temp);
    image.setSize(temp.getWidth(),temp.getHeight());
    copy(temp,image,0,0);
  }
}

void overSampledLine(Bitmap<float> &image,int x1,int y1,int x2,int y2,int n){
  int u=1<<n;
  image.setSize(u*image.getWidth(),u*image.getHeight());
  image.fill(image.black());
  for (int i=0;i<u;i++)
    for (int j=0;j<u;j++)
      line1(image,u*x1+i,u*y1+j,u*x2+i,u*y2+j);
  multiUnderSample(image,n);
}

#define X1 5
#define Y1 7
#define X2 45
#define Y2 37

int main(){
  Bitmap<float> image1(50),image2(50),image3(50),image(256,52);
  image.fill(image.red());
  line1(image1,X1,Y1,X2,Y2);
  copy(image1,image,1,1);
  line2(image2,X1,Y1,X2,Y2);
  copy(image2,image,52,1);
  for (int i=1;i<4;i++){
    overSampledLine(image3,X1,Y1,X2,Y2,i);
    copy(image3,image,52+51*i,1);
  }
  SAVE_AND_SHOW(image,"prog3.bmp");
  return 0;
}
